import datetime


def dwtimestamp_to_utc(dwtimestamp: str) -> datetime.datetime:
    dt = datetime.datetime.strptime(dwtimestamp, "%Y-%m-%d %H:%M:%S")
    dt = dt.replace(tzinfo=datetime.timezone.utc)
    return dt.astimezone(datetime.timezone.utc)



import datetime
import logging
import pathlib

import numpy
import pandas
import pytz

from .. import core
from . import common

logger = logging.getLogger("detl.parsing.dw4")


BLOCKPARSERS = {
    "Info": common.parse_generic,
    "CoreInfo": common.parse_generic,
    "ProjectInfo": common.parse_generic_T,
    "TrackInfo": common.parse_generic_T,
    "TrackData": common.parse_generic,
    "Setup": common.parse_generic,
    "Unit": common.parse_generic,
    "Requirements": common.parse_requirements,
    "Sensor Elements": common.parse_generic,
    "Device Channels": common.parse_generic,
    "Profiles": common.parse_profiles,
    "Events": common.parse_generic,
    "Fb-Pro": common.parse_generic,
    "Procedure": common.parse_generic,
    "Profile Columns": common.parse_profile_columns,
    "Plant": common.parse_generic,
    "Units": common.parse_generic,
    "Sensors": common.parse_generic,
    "Modules": common.parse_generic,
    "External Servers": common.parse_generic,
    "External Values": common.parse_generic,
    "Internal Values": common.parse_generic,
    "Setups": common.parse_generic,
}

columnmapping = {
    "volume_pv": "V\\d\\.PV",
    "temperature_sp": "T\\d\\.SP",
    "temperature_pv": "T\\d\\.PV",
    "temperature_out": "T\\d\\.Out",
    "stirrer_speed_sp": "N\\d\\.SP",
    "stirrer_speed_pv": "N\\d\\.PV",
    "stirrer_torque_pv": "Torque\\d\\.PV",
    "aeration_sp": "F\\d\\.SP",
    "aeration_pv": "F\\d\\.PV",
    "aeration_x_o2_sp": "XO2 \\d\\.SP",
    "aeration_x_o2_pv": "XO2 \\d\\.PV",
    "aeration_x_co2_sp": "XCO2 \\d\\.SP",
    "aeration_x_co2_pv": "XCO2 \\d\\.PV",
    "ph_sp": "pH\\d\\.SP",
    "ph_pv": "pH\\d\\.PV",
    "ph_out": "pH\\d\\.Out",
    "do_sp": "DO\\d\\.SP",
    "do_pv": "DO\\d\\.PV",
    "do_out": "DO\\d\\.Out",
    "off-gas_pv": "F\\d\\.Out",
    "off-gas_x_o2_pv": "XO2 \\d\\.Out",
    "off-gas_x_co2_pv": "XCO2 \\d\\.Out",
    "otr_pv": "OTR\\d",
    "vot_pv": "VOT\\d",
    "ctr_pv": "CTR\\d",
    "vct_pv": "VCT\\d",
    "rq_pv": "RQ\\d",
    "pump_a_rate_sp": "FA\\d\\.SP",
    "pump_a_rate_pv": "FA\\d\\.PV",
    "pump_a_volume_pv": "VA\\d\\.PV",
    "pump_b_rate_sp": "FB\\d\\.SP",
    "pump_b_rate_pv": "FB\\d\\.PV",
    "pump_b_volume_pv": "VB\\d\\.PV",
    "pump_c_rate_sp": "FC\\d\\.SP",
    "pump_c_rate_pv": "FC\\d\\.PV",
    "pump_c_volume_pv": "VC\\d\\.PV",
    "pump_d_rate_sp": "FD\\d\\.SP",
    "pump_d_rate_pv": "FD\\d\\.PV",
    "pump_d_volume_pv": "VD\\d\\.PV",
    "absorption_pv": "AU\\d",
    "turbidity_pv": "CX\\d",
    "level_pv": "Level\\d\\.PV",
    "redox_sp": "RD\\d\\.SP",
    "redox_pv": "RD\\d\\.PV",
    "redox_out": "RD\\d\\.Out",
    "offline_a": "Offline\\d\\.A",
    "offline_b": "Offline\\d\\.B",
    "offline_c": "Offline\\d\\.C",
    "offline_d": "Offline\\d\\.D",
    "opc_a": "OPC\\d\\.A",
    "opc_b": "OPC\\d\\.B",
    "opc_c": "OPC\\d\\.C",
    "opc_d": "OPC\\d\\.D",
    "balance_a_p": "MA\\d\\.PV",
    "balance_b_pv": "MB\\d\\.PV",
    "balance_c_pv": "MC\\d\\.PV",
    "balance_d_pv": "MD\\d\\.PV",
    "external_a_pv": "External\\d\\.A",
    "external_b_pv": "External\\d\\.B",
    "external_c_pv": "External\\d\\.C",
    "external_d_pv": "External\\d\\.D",
}


class DASware4Parser(core.DASwareParser):
    def parse(self, filepath) -> core.DWData:
        """Parses the provided DASware CSV file into a data object.

        Args:
            filepath (str or pathlib.Path): path pointing to the file of interest
        """
        scoped_blocks = common.split_blocks(filepath)
        dd = common.transform_to_dwdata(scoped_blocks, BLOCKPARSERS, version=core.DASwareVersion.V5)

        for _, reactor in dd.items():
            reactor._dataframe = common.transform_trackdata(
                reactor.trackdata, columnmapping, core.DASwareVersion.V4
            )

        return dd



import logging
import pathlib

import numpy
import pandas

from .. import core
from . import common

logger = logging.getLogger("detl.parsing.dw5")


BLOCKPARSERS = {
    "Info": common.parse_generic,
    "CoreInfo": common.parse_generic,
    "ProjectInfo": common.parse_generic_T,
    "TrackInfo": common.parse_generic_T,
    "TrackData": common.parse_generic,
    "Setup": common.parse_generic,
    "Unit": common.parse_generic,
    "Requirements": common.parse_requirements,
    "Sensor Elements": common.parse_generic,
    "Device Channels": common.parse_generic,
    "Profiles": common.parse_profiles,
    "Events": common.parse_generic,
    "Fb-Pro": common.parse_generic,
    "Procedure": common.parse_generic,
    "Profile Columns": common.parse_profile_columns,
    "Plant": common.parse_generic,
    "Units": common.parse_generic,
    "Sensors": common.parse_generic,
    "Modules": common.parse_generic,
    "External Servers": common.parse_generic,
    "External Values": common.parse_generic,
    "Internal Values": common.parse_generic,
    "Setups": common.parse_generic,
}

columnmapping = {
    "volume_pv": "\\.V\\d+\\.VPV",
    "temperature_sp": "\\.T\\d+\\.SP",
    "temperature_pv": "\\.\\.T\\d+\\.PV",
    "temperature_out": "\\.T\\d+\\.Out",
    "stirrer_speed_sp": "\\.N\\d+\\.SP",
    "stirrer_speed_pv": "\\.N\\d+\\.PV",
    "stirrer_torque_pv": "\\.N\\d+\\.TStirPV",
    "aeration_sp": "\\.F\\d+\\.SP",
    "aeration_pv": "\\.F\\d+\\.PV",
    "aeration_air_pv": "\\.FAir\\d+\\.PV",
    "aeration_co2_pv": "\\.FCO2\\d+\\.PV",
    "aeration_o2_pv": "\\.FO2\\d+\\.PV",
    "aeration_n2_pv": "\\.FN2\\d+\\.PV",
    "aeration_x_co2_pv": "\\.XCO2\\d+\\.PV",
    "aeration_x_o2_pv": "\\.XO2\\d+\\.PV",
    "ph_sp": "\\.pH\\d+\\.SP",
    "ph_pv": "\\.pH\\d+\\.PV",
    "ph_out": "\\.pH\\d+\\.Out",
    "do_sp": "\\.DO\\d+\\.SP",
    "do_pv": "\\.DO\\d+\\.PV",
    "do_out": "\\.DO\\d+\\.Out",
    "off-gas_pv": "\\.F\\d+\\.Out",
    "off-gas_x_o2_pv": "\\.XO2\\d+\\.Out",
    "off-gas_x_co2_pv": "\\.XCO2\\d+\\.Out",
    "otr_pv": "\\.OTR\\d+",
    "vot_pv": "\\.VOT\\d+",
    "ctr_pv": "\\.CTR\\d+",
    "vct_pv": "\\.VCT\\d+",
    "rq_pv": "\\.RQ\\d+",
    "pump_a_rate_sp": "\\.FA\\d+\\.SP",
    "pump_a_rate_pv": "\\.FA\\d+\\.PV",
    "pump_a_volume_pv": "\\.VA\\d+\\.PV",
    "pump_b_rate_sp": "\\.FB\\d+\\.SP",
    "pump_b_rate_pv": "\\.FB\\d+\\.PV",
    "pump_b_volume_pv": "\\.VB\\d+\\.PV",
    "pump_c_rate_sp": "\\.FC\\d+\\.SP",
    "pump_c_rate_pv": "\\.FC\\d+\\.PV",
    "pump_c_volume_pv": "\\.VC\\d+\\.PV",
    "pump_d_rate_sp": "\\.FD\\d+\\.SP",
    "pump_d_rate_pv": "\\.FD\\d+\\.PV",
    "pump_d_volume_pv": "\\.VD\\d+\\.PV",
    "absorption_pv": "\\.ODAU\\d+\\.PV",
    "turbidity_pv": "\\.ODCX\\d+\\.PV",
    "level_pv": "\\.Lvl\\d+\\.PV",
    "offline_a": "\\.OfflineA\\d+\\.OfflineA",
    "offline_b": "\\.OfflineB\\d+\\.OfflineB",
    "offline_c": "\\.OfflineC\\d+\\.OfflineC",
    "offline_d": "\\.OfflineD\\d+\\.OfflineD",
    "offline_e": "\\.OfflineE\\d+\\.OfflineE",
    "offline_f": "\\.OfflineF\\d+\\.OfflineF",
    "offline_g": "\\.OfflineG\\d+\\.OfflineG",
    "offline_h": "\\.OfflineH\\d+\\.OfflineH",
    "offline_i": "\\.OfflineI\\d+\\.OfflineI",
    "offline_j": "\\.OfflineJ\\d+\\.OfflineJ",
    "offline_k": "\\.OfflineK\\d+\\.OfflineK",
    "offline_l": "\\.OfflineL\\d+\\.OfflineL",
    "offline_m": "\\.OfflineM\\d+\\.OfflineM",
    "offline_n": "\\.OfflineN\\d+\\.OfflineN",
    "offline_o": "\\.OfflineO\\d+\\.OfflineO",
    "offline_p": "\\.OfflineP\\d+\\.OfflineP",
    "offline_q": "\\.OfflineQ\\d+\\.OfflineQ",
    "offline_r": "\\.OfflineR\\d+\\.OfflineR",
    "offline_s": "\\.OfflineS\\d+\\.OfflineS",
    "offline_t": "\\.OfflineT\\d+\\.OfflineT",
    "offline_u": "\\.OfflineU\\d+\\.OfflineU",
    "offline_v": "\\.OfflineV\\d+\\.OfflineV",
    "offline_w": "\\.OfflineW\\d+\\.OfflineW",
    "offline_x": "\\.OfflineX\\d+\\.OfflineX",
    "offline_y": "\\.OfflineY\\d+\\.OfflineY",
    "offline_z": "\\.OfflineZ\\d+\\.OfflineZ",
    "loop_a_sp": "\\.LoopA\\d+\\.SP",
    "loop_a_pv": "\\.LoopA\\d+\\.PV",
    "loop_a_out": "\\.LoopA\\d+\\.Out",
    "loop_b_sp": "\\.LoopB\\d+\\.SP",
    "loop_b_pv": "\\.LoopB\\d+\\.PV",
    "loop_b_out": "\\.LoopB\\d+\\.Out",
    "loop_c_sp": "\\.LoopC\\d+\\.SP",
    "loop_c_pv": "\\.LoopC\\d+\\.PV",
    "loop_c_out": "\\.LoopC\\d+\\.Out",
    "loop_d_sp": "\\.LoopD\\d+\\.SP",
    "loop_d_pv": "\\.LoopD\\d+\\.PV",
    "loop_d_out": "\\.LoopD\\d+\\.Out",
    "internal_a_start_oszillation": "\\.InternalA\\d+\\.Start Oszillation",
}


class DASware5Parser(core.DASwareParser):
    def parse(self, filepath: pathlib.Path) -> core.DWData:
        """Parses the provided DASware CSV file into a data object.

        Args:
            filepath (str or pathlib.Path): path pointing to the file of interest
        """
        scoped_blocks = common.split_blocks(filepath)
        scoped_blocks = {
            key: value for (key, value) in scoped_blocks.items() if "TrackData" in list(value)
        }
        dd = common.transform_to_dwdata(scoped_blocks, BLOCKPARSERS, version=core.DASwareVersion.V5)

        for _, reactor in dd.items():
            reactor._dataframe = common.transform_trackdata(
                reactor.trackdata, columnmapping, core.DASwareVersion.V5
            )

        return dd



"""Contains unit tests for the `detl` package"""
import datetime
import pathlib
import unittest

import numpy
import pandas

import detl

dir_testfiles = pathlib.Path(pathlib.Path(__file__).absolute().parent, "testfiles")

v4_testfiles = [
    pathlib.Path(dir_testfiles, "v4_NT-WMB-2.Control.csv"),
    pathlib.Path(dir_testfiles, "v4_20180726.Control.csv"),
]
v4_nreactors = [4, 4]
v4_trackdata_nrows = [
    [5461, 5460, 5460, 5459],
    [1371, 1370, 1370, 1370],
]

v5_testfiles = [
    pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv"),
    pathlib.Path(dir_testfiles, "v5_medium_CTPC06110.20181219.Control.csv"),
    pathlib.Path(dir_testfiles, "v5_long_CTPC06110.20190218.Control.csv"),
]
v5_nreactors = [
    4,
    4,
    4,
]
v5_trackdata_nrows = [
    [14111, 14112, 14108, 14112],
    [16936, 16936, 16935, 16936],
    [52245, 52245, 52245, 52245],
]


class TestParserSelection(unittest.TestCase):
    def test_v4_detection(self):
        parser = detl.get_parser(pathlib.Path(dir_testfiles, "v4_20180726.Control.csv"))
        self.assertIsNotNone(parser)
        self.assertIsInstance(parser, detl.core.DASwareParser)
        self.assertIsInstance(parser, detl.parsing.dw4.DASware4Parser)
        return

    def test_v5_detection(self):
        parser = detl.get_parser(pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv"))
        self.assertIsNotNone(parser)
        self.assertIsInstance(parser, detl.core.DASwareParser)
        self.assertIsInstance(parser, detl.parsing.dw5.DASware5Parser)
        return

    def test_invalid_detection(self):
        with self.assertRaises(NotImplementedError):
            _ = detl.get_parser(pathlib.Path(dir_testfiles, "invalid.csv"))
        return


class TestCommonParsing(unittest.TestCase):
    def test_split_blocks_v5(self):
        with self.assertRaises(ValueError):
            detl.parsing.common.split_blocks(["bla"])

        filepath = pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv")

        scoped_blocks = detl.parsing.common.split_blocks(filepath)
        self.assertEqual(len(scoped_blocks), 5)
        self.assertTrue(None in scoped_blocks)
        self.assertTrue(1 in scoped_blocks)
        self.assertTrue(2 in scoped_blocks)
        self.assertTrue(3 in scoped_blocks)
        self.assertTrue(4 in scoped_blocks)
        self.assertTrue("TrackData" not in scoped_blocks)
        self.assertTrue("Events" in scoped_blocks[None])
        self.assertTrue("Fb-Pro" in scoped_blocks[None])
        self.assertTrue("Procedure" in scoped_blocks[None])
        self.assertTrue("Profile Columns" in scoped_blocks[None])
        self.assertTrue("Plant" in scoped_blocks[None])
        self.assertTrue("Units" in scoped_blocks[None])
        self.assertTrue("Sensors" in scoped_blocks[None])
        self.assertTrue("Modules" in scoped_blocks[None])
        self.assertTrue("External Servers" in scoped_blocks[None])
        self.assertTrue("External Values" in scoped_blocks[None])
        self.assertTrue("Internal Values" in scoped_blocks[None])
        self.assertTrue("Setups" in scoped_blocks[None])
        for r in {1, 2, 3, 4}:
            self.assertTrue("TrackData" in scoped_blocks[r])
            self.assertTrue("Setup" in scoped_blocks[r])
            self.assertTrue("Unit" in scoped_blocks[r])
            self.assertTrue("Requirements" in scoped_blocks[r])
            self.assertTrue("Sensor Elements" in scoped_blocks[r])
            self.assertTrue("Device Channels" in scoped_blocks[r])
            self.assertTrue("Profiles" in scoped_blocks[r])
            self.assertTrue("Profiles" in scoped_blocks[r])
        return

    def test_parse_generic(self):
        filepath = pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv")
        scoped_blocks = detl.parsing.common.split_blocks(filepath)
        attr, df = detl.parsing.common.parse_generic(
            "Info", scoped_blocks[None]["Info"], scope=None
        )
        self.assertEqual(attr, "_info")
        self.assertIsInstance(df, pandas.DataFrame)
        self.assertEqual(len(df), 71)
        self.assertEqual(len(df.columns), 7)

    def test_parse_generic_T(self):
        filepath = pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv")
        scoped_blocks = detl.parsing.common.split_blocks(filepath)
        attr, df = detl.parsing.common.parse_generic_T(
            "Info", scoped_blocks[None]["Info"], scope=None
        )
        self.assertEqual(attr, "_info")
        self.assertIsInstance(df, pandas.DataFrame)
        self.assertEqual(len(df), 7)
        self.assertEqual(len(df.columns), 71)

    def test_transform_to_dwdata_v5(self):
        filepath = pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv")
        scoped_blocks = detl.parsing.common.split_blocks(filepath)
        dd = detl.parsing.common.transform_to_dwdata(
            scoped_blocks, detl.parsing.dw5.BLOCKPARSERS, detl.DASwareVersion.V5
        )
        self.assertIsInstance(dd, detl.DWData)
        self.assertIsInstance(dd, dict)
        self.assertEqual(dd.version, detl.DASwareVersion.V5)

    def test_inoculation_times(self):
        filepath = pathlib.Path(dir_testfiles, "v4_NT-WMB-2.Control.csv")
        dd = detl.parse(
            filepath,
            inoculation_times={
                1: datetime.datetime(2016, 3, 9, 16, 38, 31, tzinfo=datetime.timezone.utc),
                2: datetime.datetime(2016, 3, 9, 16, 50, 31, tzinfo=datetime.timezone.utc),
            },
        )
        self.assertEqual(dd[1].dataframe.process_time[0], 0)
        self.assertTrue(numpy.isnan(dd[3].dataframe.process_time[0]))
        self.assertTrue(numpy.isnan(dd[2].dataframe.process_time[0]))
        return


class TestDW4Parsing(unittest.TestCase):
    def test_trackdata_row_count(self):
        for fp, row_counts, nunits in zip(v4_testfiles, v4_trackdata_nrows, v4_nreactors):
            ddata = detl.parse(fp)
            for r, nrows in zip(range(1, nunits + 1), row_counts):
                self.assertTrue(r in ddata)
                self.assertIsNotNone(ddata[r].trackdata)
                self.assertEqual(len(ddata[r].trackdata), nrows)
        return

    def test_trackdata_transformation(self):
        ddata = detl.parse(v4_testfiles[0])

        assert "pump_a_rate_sp" in ddata[1].dataframe.columns
        assert "pump_b_rate_sp" in ddata[1].dataframe.columns

        self.assertAlmostEqual(ddata[1].dataframe.loc[2798, "duration"], 23.325, places=3)
        self.assertAlmostEqual(ddata[1].dataframe.loc[2798, "process_time"], 8.74667, places=3)
        self.assertAlmostEqual(ddata[2].dataframe.loc[4513, "off-gas_pv"], 31.675, places=3)
        self.assertAlmostEqual(ddata[3].dataframe.loc[1472, "temperature_pv"], 30.011, places=3)
        self.assertAlmostEqual(ddata[4].dataframe.loc[3475, "pump_a_volume_pv"], 0.622, places=3)
        self.assertEqual(
            ddata[4].dataframe.loc[2878, "timestamp"],
            datetime.datetime(2016, 3, 10, 16, 38, 8, tzinfo=datetime.timezone.utc),
        )

    def test_timestamp_parsing(self):
        ddata = detl.parse(pathlib.Path(dir_testfiles, "v4_NT-WMB-2.Control.csv"))
        self.assertEqual(
            ddata[1].dataframe.loc[0, "timestamp"],
            datetime.datetime(2016, 3, 9, 16, 38, 31, tzinfo=datetime.timezone.utc),
        )

        ddata = detl.parse(pathlib.Path(dir_testfiles, "v4_20180726.Control.csv"))
        self.assertEqual(
            ddata[1].dataframe.loc[0, "timestamp"],
            datetime.datetime(2018, 7, 26, 11, 53, 36, tzinfo=datetime.timezone.utc),
        )

        return


class TestDW5Parsing(unittest.TestCase):
    def test_trackdata_row_count(self):
        for fp, row_counts, nunits in zip(v5_testfiles, v5_trackdata_nrows, v5_nreactors):
            ddata = detl.parse(fp)
            for r, nrows in zip(range(1, nunits + 1), row_counts):
                self.assertTrue(r in ddata)
                self.assertIsNotNone(ddata[r].trackdata)
                self.assertEqual(len(ddata[r].trackdata), nrows)
        return

    def test_trackdata_transformation(self):
        ddata = detl.parse(v5_testfiles[1])

        assert "pump_a_rate_sp" in ddata[1].dataframe.columns
        assert "pump_b_rate_sp" in ddata[1].dataframe.columns

        self.assertAlmostEqual(ddata[1].dataframe.loc[11359, "aeration_x_co2_pv"], 0.034, places=3)
        self.assertAlmostEqual(ddata[2].dataframe.loc[4128, "stirrer_speed_pv"], 1059.382, places=3)
        self.assertAlmostEqual(ddata[3].dataframe.loc[13387, "duration"], 37.238, places=3)
        self.assertAlmostEqual(ddata[3].dataframe.loc[13387, "process_time"], 36.3028, places=3)
        self.assertAlmostEqual(ddata[4].dataframe.loc[9400, "do_sp"], 30.0, places=3)

    def test_timestamp_parsing(self):
        ddata = detl.parse(pathlib.Path(dir_testfiles, "v5_short_CTPC06280.Control.csv"))
        self.assertEqual(
            ddata[1].dataframe.loc[0, "timestamp"],
            datetime.datetime(2019, 2, 6, 10, 46, 52, tzinfo=datetime.timezone.utc),
        )
        return


class TestClosestDataLookup(unittest.TestCase):
    def test_dw4(self):
        ddata = detl.parse(v4_testfiles[0])
        points = numpy.array([0, 5, 15, 1999])
        expected_output = numpy.array([1000.027, 1003.670, 1026.961, 1034.168])

        looked_up_data = ddata[1].get_closest_data(points)
        [
            self.assertAlmostEqual(i, o, places=3)
            for i, o in zip(looked_up_data["volume_pv"], expected_output)
        ]

    def test_dw5(self):
        ddata = detl.parse(v5_testfiles[0])
        points = numpy.array([0, 4, 16, 100])
        expected_output = numpy.array([0, 0.291, 5.481, 26.343])

        looked_up_data = ddata[4].get_closest_data(points, reference="duration")
        [
            self.assertAlmostEqual(i, o, places=3)
            for i, o in zip(looked_up_data["ctr_pv"], expected_output)
        ]


class TestGetNarrowData(unittest.TestCase):
    def test_with_process_time(self):
        ddata = detl.parse(v4_testfiles[0])
        nd = ddata.get_narrow_data()
        self.assertAlmostEqual(nd.loc[5643]["value"], 1030.005, places=3)
        self.assertAlmostEqual(nd.loc[25146]["time"], 23.8967, places=3)

    def test_with_duration(self):
        ddata = detl.parse(v4_testfiles[0])
        nd = ddata.get_narrow_data(kdim="duration")
        self.assertAlmostEqual(nd.loc[20456]["value"], 29.889, places=3)
        self.assertAlmostEqual(nd.loc[15643]["time"], 39.35, places=3)

    def test_kdim_setting(self):
        ddata = detl.parse(v4_testfiles[0])
        with self.assertRaises(KeyError):
            ddata.get_narrow_data(kdim="volume_pv")


class TestReactorDataProps(unittest.TestCase):
    def test_reactor_data(self):
        ddata = detl.parse(v4_testfiles[0])
        rdata = list(ddata.values())[0]
        self.assertIsInstance(rdata, detl.core.ReactorData)

        # check properties
        rdata.id
        rdata.setup
        rdata.unit
        rdata.requirements
        rdata.sensor_elements
        rdata.device_channels
        rdata.profiles
        rdata.trackdata
        rdata.dataframe


if __name__ == "__main__":
    unittest.main()



"""Specifies the base types for parsing and representing DASware data."""
import abc
import enum
import pathlib
from typing import Dict

import numpy
import pandas


class DASwareVersion(enum.Enum):
    V4 = "v4"
    V5 = "v5"


class ReactorData(object):
    """Data structure containing data from one reactor."""

    def __init__(self, id: int):
        self._id = id
        self._setup = None
        self._unit = None
        self._requirements = None
        self._sensor_elements = None
        self._device_channels = None
        self._profiles = None
        self._trackdata = None
        self._dataframe = None

    @property
    def id(self) -> int:
        """Number of the reactor."""
        return self._id

    @property
    def setup(self) -> pandas.DataFrame:
        """Dataframe of overall process information."""
        return self._setup

    @property
    def unit(self) -> pandas.DataFrame:
        """Properties of the reactor."""
        return self._unit

    @property
    def requirements(self) -> pandas.DataFrame:
        return self._requirements

    @property
    def sensor_elements(self) -> pandas.DataFrame:
        """Table of connected sensors."""
        return self._sensor_elements

    @property
    def device_channels(self) -> pandas.DataFrame:
        return self._device_channels

    @property
    def profiles(self) -> pandas.DataFrame:
        return self._profiles

    @property
    def trackdata(self) -> pandas.DataFrame:
        """Contains timeseries of mass flows."""
        return self._trackdata

    @property
    def dataframe(self) -> pandas.DataFrame:
        """Primary table of setpoint (SP) and actual (PV) control parameters."""
        return self._dataframe

    def get_closest_data(
        self, points: numpy.array, reference: str = "process_time"
    ) -> pandas.DataFrame:
        """Returns a subset of the reactor data at points closest to the given ones.

        Args:
            points (numpy.array): the data from readings closest to these points will be returned
            reference (str): name of the column to look for points

        Returns:
            filtered_data: DataFrame containing data closest to the given points

        Raises:
            KeyError: when the reference column is not in the DataFrame
        """
        if reference not in self.dataframe.columns:
            raise KeyError("Reference column not in DataFrame")

        idx = [abs(self.dataframe.loc[:, reference] - p).idxmin() for p in points]

        return self.dataframe.loc[idx]


class DWData(Dict[str, ReactorData]):
    """Standardized data type for DASGIP data."""

    def __init__(self, version: DASwareVersion):
        self._version = version
        self._info = None
        self._coreinfo = None
        self._projectinfo = None
        self._trackinfo = None
        self._events = None
        self._fb_pro = None
        self._procedure = None
        self._profile_columns = None
        self._plant = None
        self._units = None
        self._sensors = None
        self._modules = None
        self._external_servers = None
        self._external_values = None
        self._internal_values = None

    @property
    def version(self) -> DASwareVersion:
        """Specifies which DASWARE version was used."""
        return self._version

    @property
    def info(self) -> pandas.DataFrame:
        """Contains version numbers of software modules."""
        return self._info

    @property
    def coreinfo(self) -> pandas.DataFrame:
        """Information about DASWARE and timezone settings."""
        return self._coreinfo

    @property
    def projectinfo(self) -> pandas.DataFrame:
        return self._projectinfo

    @property
    def tracks(self) -> pandas.DataFrame:
        """Metadata about logging settings."""
        return self._trackinfo

    @property
    def events(self) -> pandas.DataFrame:
        """Table of events that happened during the process."""
        return self._events

    @property
    def fb_pro(self) -> pandas.DataFrame:
        return self._fb_pro

    @property
    def procedure(self) -> pandas.DataFrame:
        """Metadata of the experiment"""
        return self._procedure

    @property
    def profile_columns(self) -> pandas.DataFrame:
        return self._profile_columns

    @property
    def plant(self) -> pandas.DataFrame:
        """Metadata of the hardware"""
        return self._plant

    @property
    def units(self) -> pandas.DataFrame:
        """Metadata of the reactor units"""
        return self._units

    @property
    def sensors(self) -> pandas.DataFrame:
        """Metadata of the connected sensors"""
        return self._sensors

    @property
    def modules(self) -> pandas.DataFrame:
        return self._modules

    @property
    def external_servers(self) -> pandas.DataFrame:
        return self._external_servers

    @property
    def external_values(self) -> pandas.DataFrame:
        return self._external_values

    @property
    def internal_values(self) -> pandas.DataFrame:
        return self._internal_values

    def get_narrow_data(self, kdim: str = "process_time"):
        """Returns all data in a narrow DataFrame.

        Args:
            kdim (str): name of the time axis which will be the time axis in the new format.
                        Can be 'timestamp', 'duration', or 'process_time'. 'process_time' will
                        drop all data before valid time information is available.
                        'duration' and 'process_time' will drop the timestamp information to get a clean
                        float-type value column.

        Returns:
            narrow_data: DataFrame containing all data in a narrow format

        Raises:
            KeyError: when the reference column is not in the DataFrame
        """
        if kdim not in {"timestamp", "duration", "process_time"}:
            raise KeyError("kdim must be 'timestamp', 'duration', or 'process_time'.")

        narrow_data = pandas.DataFrame()
        for reactor_id in self:
            if kdim == "process_time":
                _df = (
                    self[reactor_id]
                    .dataframe.dropna()
                    .drop("timestamp", axis="columns")
                    .melt(id_vars=kdim)
                )
            elif kdim == "duration":
                _df = (
                    self[reactor_id].dataframe.drop("timestamp", axis="columns").melt(id_vars=kdim)
                )
            else:
                _df = self[reactor_id].dataframe.melt(id_vars=kdim)

            _df["reactor"] = reactor_id
            _df = _df.astype({"value": float})
            narrow_data = pandas.concat([narrow_data, _df], ignore_index=True)

        narrow_data = narrow_data.rename({kdim: "time"}, axis="columns")[
            ["reactor", "time", "variable", "value"]
        ]

        return narrow_data


class DASwareParser(object):
    """Abstract type for parsers that read DASware CSV files."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def parse(self, filepath: pathlib.Path) -> DWData:
        """Parses the provided DASware CSV file into a data object.

        Args:
            filepath (str or pathlib.Path): path pointing to the file of interest
        """
        raise NotImplementedError(
            "Whoever implemented {} screwed up.".format(self.__class__.__name__)
        )



import pathlib

import setuptools

__packagename__ = "detl"
ROOT = pathlib.Path(__file__).parent


def get_version():
    import os
    import re

    VERSIONFILE = os.path.join(__packagename__, "__init__.py")
    initfile_lines = open(VERSIONFILE, "rt").readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


__version__ = get_version()


setuptools.setup(
    name=__packagename__,
    packages=setuptools.find_packages(),  # this must be the same as the name above
    version=__version__,
    description="Package for parsing and transforming DASware raw data exports.",
    url="https://github.com/JuBiotech/detl",
    author="IBG-1",
    copyright="(c) 2022 Forschungszentrum Jülich GmbH",
    license="GNU Affero General Public License v3.0",
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ],
    install_requires=open(pathlib.Path(ROOT, "requirements.txt")).readlines(),
)



import collections
import datetime
import logging
import pathlib
import re
import warnings
from io import StringIO

import numpy
import pandas

from .. import core
from . import utils

logger = logging.getLogger("detl.parsing.common")


def split_blocks(filepath: pathlib.Path) -> dict:
    """Reads a CSV file and splits its contents into scoped blocks.

    Args:
        filepath (pathlib.Path): path to the raw CSV

    Returns:
        scoped_blocks (dict): dicationary mapping scope to dictionary of blocks
    """
    if not isinstance(filepath, (str, pathlib.Path)):
        raise ValueError("Please provide filepath either as str or pathlib.Path object")

    # split the entire file into table-blocks
    blocks = [[]]
    with pathlib.Path(filepath).open(mode="r", errors="replace") as file:
        for line in file:
            if len(line) == 1:
                blocks.append([])
            else:
                blocks[-1].append(line)
    # drop empty blocks
    blocks = [block for block in blocks if len(block) > 1]

    # group blocks by scope (None or reactor-number)
    scoped_blocks = collections.defaultdict(dict)
    scope = None
    for blocklines in blocks:
        blockheader = blocklines[0].strip()
        setup_matches = re.findall(r'"\[Setup(\d+)\]"', blockheader)
        track_matches = re.findall(r'"\[TrackData(\d+)\]"', blockheader)
        if len(track_matches) == 1:
            scope = int(track_matches[0])
        elif blockheader == '"[Events]"':
            scope = None
        elif len(setup_matches) == 1:
            scope = int(setup_matches[0])
        blockheader = blockheader[2:-2]
        if scope:
            blockheader = blockheader.strip(str(scope))
        scoped_blocks[scope][blockheader] = "".join(blocklines[1:]).strip()
    return scoped_blocks


def transform_to_dwdata(
    scoped_blocks: dict, blockparsers: dict, version: core.DASwareVersion
) -> core.DWData:
    dd = core.DWData(version)
    for scope, blocks in scoped_blocks.items():
        if scope is not None and not scope in dd:
            dd[scope] = core.ReactorData(scope)
        for header, block in blocks.items():
            if not header in blockparsers:
                logger.warn(f'No parser found for block "{header}"')
                continue
            blockparser = blockparsers[header]
            if blockparser is not None:
                try:
                    attr, df = blockparser(header, block, scope)
                    if scope is None:
                        setattr(dd, attr, df)
                    else:
                        setattr(dd[scope], attr, df)
                except NotImplementedError as ex:
                    logger.debug(
                        f'scope {scope}: Parsing function for "{header}" is not implemented'
                    )
                except:
                    logger.warning(f'scope {scope}: Failed to parse block "{header}"')
    return dd


def parse_generic(header, block, scope):
    df = pandas.read_csv(StringIO(block), sep=";")
    attr = "_" + header.lower().replace(" ", "_").replace("-", "_")
    return (attr, df)


def parse_generic_T(header, block, scope):
    attr, df = parse_generic(header, block, scope)
    return (attr, df.T)


def parse_requirements(header, block, scope):
    raise NotImplementedError()


def parse_profiles(header, block, scope):
    raise NotImplementedError()


def parse_profile_columns(header, block, scope):
    raise NotImplementedError()


def transform_trackdata(
    trackdata: pandas.DataFrame, columnmapping: dict, version: core.DASwareVersion
) -> pandas.DataFrame:
    """Parses trackdata to an useful DataFrame.

    Args:
        trackdata (pandas.DataFrame): Trackdata derived from DASGIP raw data file
        columnmapping (dict): Mapping from trackdata column names to reasonable column names
        version (core.DASwareVersion): inform about the DASware version of the file that is being processed

    Returns:
        transformed_data (pandas.DataFrame): DataFrame with structured data
    """
    transformed_data = pandas.DataFrame(
        index=trackdata.index,
        columns=["timestamp", "duration", "process_time"],
    )
    transformed_data["timestamp"] = [utils.dwtimestamp_to_utc(t) for t in trackdata["Timestamp"]]
    transformed_data["duration"] = trackdata["Duration"] * 24

    magic_time = datetime.datetime.strptime("1899-12-30 00:00:00", "%Y-%m-%d %H:%M:%S")
    switch = False
    if version == core.DASwareVersion.V4:
        ser = trackdata.filter(regex=".*Inoculation Time.*", axis="columns").squeeze()
    elif version == core.DASwareVersion.V5:
        ser = trackdata.filter(regex=".*InoculationTime.*", axis="columns").squeeze()
    else:
        raise NotImplementedError(f"Unknown DASwareVersion: {version}")
    process_time = numpy.full(len(ser), numpy.nan)

    if not ser.empty:
        for i in range(len(ser)):
            if pandas.notna(ser[i]):
                td = datetime.datetime.strptime(ser[i], "%Y-%m-%d %H:%M:%S") - magic_time

                if (not switch) and (td.total_seconds() > 0):
                    switch = True
                    process_time[i - 1] = float(0)

                if switch:
                    process_time[i] = td.total_seconds() / 3600

    transformed_data["process_time"] = process_time

    for key, reg in columnmapping.items():
        new_data = trackdata.filter(regex=reg, axis="columns").squeeze()
        if (not new_data.empty) and (not new_data.isnull().all()):
            transformed_data.loc[:, key] = new_data

    transformed_data = transformed_data.fillna(method="ffill")

    return transformed_data



from . import parsing
from .core import DASwareParser, DASwareVersion, DWData

__version__ = "1.1.2"

parsers = {
    DASwareVersion.V4: parsing.dw4.DASware4Parser,
    DASwareVersion.V5: parsing.dw5.DASware5Parser,
}


def get_parser(filepath) -> DASwareParser:
    """Analyzes a raw DASware CSV file and selects an appropiate parser.

    Args:
        filepath (str or pathlib.Path): path pointing to the file of interest

    Returns:
        DWDParser: a parser that can be used for the provided file type

    Raises:
        NotImlementedError: when the file contents do not match with a known DASware CSV style
    """
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            lines = f.readlines()
    except UnicodeDecodeError:
        with open(filepath, "r", encoding="latin-1") as f:
            lines = f.readlines()

    version = None

    if len(lines) > 2 and lines[2].startswith('"FngArchiv";"4.0.1"'):
        version = DASwareVersion.V4
    elif len(lines) > 2 and lines[2].startswith('"FngArchiv";"5.0.0"'):
        version = DASwareVersion.V5
    else:
        raise NotImplementedError("Unsupported file version")

    # select a parser for this version
    parser_cls = parsers[version]
    return parser_cls()


def parse(filepath, *, inoculation_times: dict = None) -> DWData:
    """Parses a raw DASware CSV file into a DWData object.

    Args:
        filepath (str or pathlib.Path): path pointing to the file of interest
        inoculation_times (dict or None): optional overrides for inoculation timepoints
            key (int): reactor number
            value (datetime.datetime): timezone-aware datetime object of the real inoculation time (computer clock!)

    Returns:
        DWData: parsed data object

    Raises:
        NotImlementedError: when the file contents do not match with a known DASware CSV style
    """
    parser = get_parser(filepath)
    data = parser.parse(filepath)

    if inoculation_times:
        for r, dt_inoculate in inoculation_times.items():
            new_process_time = (
                data[r].dataframe["timestamp"] - dt_inoculate
            ).dt.total_seconds() / 3600
            new_process_time[new_process_time < 0] = float("nan")
            data[r].dataframe["process_time"] = new_process_time

    return data
